from math import pi

text = f'The value of Pi is approx. {pi}'

def foo(a: str = None):
    return text

# multi line
# comment

def main():
    print(foo())  # comment


class A:
    a = 123

    def bar(self):
        return foo()


def decorate(f):
    return f


@decorate
def baz():
    pass


if __name__ == '__main__':
    def cond():
        pass

    main()



from os import environ
from pathlib import Path
from platform import machine, system
from subprocess import check_call

from .api_client import download_resource


MACHINE = machine().lower()
if MACHINE == 'x86_64':  MACHINE = 'amd64'
if MACHINE == 'aarch64':  MACHINE = 'arm64'
SYSTEM = system().lower()
BINARY_KEY = f'goscan-{SYSTEM}-{MACHINE}'


class Lang:
    def prepare_package(self, package):
        self.package = package
        self.uim_dir = package.temp_dir / 'uim'
        self._run_go_scanner(package.repo_root, self.uim_dir, package.index_system)

    def add_to_tar_file(self, package, output):
        # for uim in self.uim_dir.glob('*'):
        output.add(self.uim_dir, arcname=package.repo_root / '.territory/uim')

    def add_to_meta(self, meta):
        meta['lang'] = 'go'

    def _get_go_scanner(self):
        print('getting parser binary for platform:', BINARY_KEY)
        if SYSTEM == 'windows':
            bin_path = self.package.temp_dir / 'goscan.exe'
        else:
            bin_path = self.package.temp_dir / 'goscan'
        download_resource(
            upload_token=self.package.upload_token,
            resource=BINARY_KEY,
            destination=bin_path)
        if not SYSTEM == 'windows':
            bin_path.chmod(0o700)
        return bin_path

    def _run_go_scanner(self, scan_dir: Path, uim_output_dir: Path, system: bool):
        scanner_path = environ.get('GOSCAN_PATH')
        if not scanner_path:
            scanner_path = str(self._get_go_scanner())
        cmd = [scanner_path]
        if system:
            cmd.append('--system')
        cmd.extend([
            scan_dir,
            uim_output_dir
        ])
        check_call(cmd)



from pathlib import Path
from subprocess import check_call


class Lang:
    def prepare_package(self, package):
        self.package = package
        self.uim_dir = package.temp_dir / 'uim'
        self._run_python_scanner(package.repo_root, self.uim_dir, package.index_system)

    def add_to_tar_file(self, package, output):
        output.add(self.uim_dir, arcname=package.repo_root / '.territory/uim')

    def add_to_meta(self, meta):
        meta['lang'] = 'python'

    def _run_python_scanner(self, scan_dir: Path, uim_output_dir: Path, system: bool):
        cmd = ['python', '-m', 'territory_python_scanner']
        if system:
            cmd.append('--system')
        cmd.extend([
            str(scan_dir),
            str(uim_output_dir)
        ])
        check_call(cmd)



import json
from pathlib import Path
from shutil import which
from subprocess import check_call


EXAMPLE_REPO_DIR = Path(__file__).parent


def init_repo(test_repo_dir, lang='c', compile_commands=None):
    check_call(['cp', '-R', EXAMPLE_REPO_DIR / f'repo-{lang}', test_repo_dir])
    check_call(['git', 'init', '-b', 'main', test_repo_dir])
    check_call(['git', '-C', test_repo_dir, 'add', '.'])
    check_call(['git', '-C', test_repo_dir, 'commit', '-m', 'initial commit'])

    if lang == 'c':
        cc = which('clang')
        (test_repo_dir / 'compile_commands.json').write_text(json.dumps(compile_commands or [
            {
                'command': f'{cc} -c -o mod1.o {test_repo_dir}/mod1.c',
                'directory': str(test_repo_dir),
                'file': f'{test_repo_dir}/mod1.c'
            },
            {
                'command': f'{cc} -c -o mod2.o {test_repo_dir}/dir/mod2.c',
                'directory': str(test_repo_dir),
                'file': f'{test_repo_dir}/dir/mod2.c'
            },
        ]))



from dataclasses import dataclass
from hashlib import blake2b
from multiprocessing import cpu_count
from multiprocessing.pool import Pool
from os import environ
from pathlib import Path
from subprocess import DEVNULL, PIPE, run
import json
import re
import shlex

import tqdm

from .files import find_in_ancestors


class Lang:
    def prepare_package(self, package):
        self.compile_commands_dir = find_compile_commands_dir(package.work_dir)
        print('compilation database:', self.compile_commands_dir / 'compile_commands.json')

        self.cc_path = self.compile_commands_dir / 'compile_commands.json'
        cc_data = read_compile_commands(self.cc_path)
        cc_files = collect_details(package.temp_dir, self.compile_commands_dir, cc_data)
        package.captured_files.update(cc_files)
        self.gen_ccs_path = Path(package.temp_dir, 'compile_commands.json')
        with self.gen_ccs_path.open('w') as f:
            json.dump(cc_data, f, indent=4)

    def add_to_tar_file(self, package, output):
        output.add(self.gen_ccs_path, arcname=self.cc_path)

    def add_to_meta(self, meta):
        meta['compile_commands_dir'] = str(self.compile_commands_dir)
        meta['lang'] = 'c'


def find_compile_commands_dir(p):
    try:
        return find_in_ancestors(p, lambda p: (p / 'compile_commands.json').exists())
    except FileNotFoundError:
        raise SystemExit('no compile_commands.json found')


def read_compile_commands(cc_path):
    with cc_path.open('r') as f:
        cc_data = json.load(f)

    for cc in cc_data:
        try:
            cmd_str = cc.pop('command')
        except KeyError:
            continue
        cc['arguments'] = shlex.split(cmd_str)

    return cc_data


def collect_details(tmp_dir, cc_dir, cc_data):
    if 'CORES' in environ:
        procs = int(environ['CORES'])
    else:
        procs = cpu_count() * 2

    with \
            Pool(procs) as pool, \
            tqdm.tqdm(total=len(cc_data), desc='collecting compilation details') as progr:
        dep_paths = set()
        def _cb(details):
            idx, paths, arguments = details
            dep_paths.update(paths)
            cc_data[idx]['arguments'] = arguments
            progr.update(1)
        def _ecb(e):
            print('error:', e)
            progr.update(1)
        for i, cmd in enumerate(cc_data):
            dir_ = cmd.get('directory') or cc_dir
            p = Path(dir_, cmd['file'])
            dep_paths.add(p)
            pool.apply_async(
                _query_details,
                (i, cc_dir, tmp_dir, cmd),
                {},
                callback=_cb,
                error_callback=_ecb)
        pool.close()
        pool.join()

    return dep_paths


def remove_arg(arguments, key, count, prefix=False):
    res = []
    fi = 0
    while fi < len(arguments):
        arg = arguments[fi]

        if arg == key:
            fi += count

        elif  (prefix and arg.startswith(key)):
            fi += 1

        else:
            res.append(arg)
            fi += 1

    return res


@dataclass
class Vee:
    target: str | None = None
    angle_bracket_include_paths: list[str] | None = None


def parse_vee(text) -> Vee:
    vee = Vee()
    m = re.search('^Target: (.+)$', text, re.MULTILINE)
    if m:
        vee.target = m.group(1)

    m = re.search(r'#include <\.\.\.> search starts here:\n((^ .*\n)*)', text, re.MULTILINE)
    if m:
        vee.angle_bracket_include_paths = re.findall(r'^ (.*?)(?: \(framework directory|headermap\)\n|\n)', m.group(1), re.MULTILINE)
    return vee


def _query_details(index: int, cc_dir: Path, tmp_dir: Path, compilation_command):
    q_arguments = compilation_command['arguments'][:]

    q_arguments = remove_arg(q_arguments, '-c', 1)
    q_arguments = remove_arg(q_arguments, '-M', 1)
    q_arguments = remove_arg(q_arguments, '-MD', 1)
    q_arguments = remove_arg(q_arguments, '-MM', 1)
    q_arguments = remove_arg(q_arguments, '-MMD', 1)
    q_arguments = remove_arg(q_arguments, '-o', 2, prefix=True)
    q_arguments = remove_arg(q_arguments, '-MF', 2, prefix=True)

    deps_dir = tmp_dir / 'deps'
    deps_dir.mkdir(parents=True, exist_ok=True)
    deps_file = deps_dir / (blake2b(compilation_command['file'].encode()).hexdigest() + '.d')

    q_arguments = [q_arguments[0], '-E', '-MD', '-MF' + str(deps_file), *q_arguments[1:], '-v', '-o', '/dev/null', '-Wno-error']
    completion = run(
        q_arguments,
        cwd=compilation_command.get('directory') or cc_dir,
        stderr=PIPE,
        stdin=DEVNULL,
        text=True)

    arguments = compilation_command['arguments'][:]
    vd = parse_vee(completion.stderr)
    if vd.target is not None:
        arguments[1:1] = ['-target', vd.target]

    if vd.angle_bracket_include_paths:
        arguments = remove_arg(arguments, '-I', 2, prefix=True)
        arguments = remove_arg(arguments, '--include-directory', 2)
        arguments = remove_arg(arguments, '--include-directory=', 1, prefix=True)
        arguments = remove_arg(arguments, '-cxx-isystem', 2, prefix=True)
        arguments = remove_arg(arguments, '-ibuiltininc', 1)
        arguments = remove_arg(arguments, '-iframework', 2, prefix=True)
        arguments = remove_arg(arguments, '-iframeworkwithsysroot', 2, prefix=True)
        arguments = remove_arg(arguments, '--stdlib++-isystem', 2, prefix=True)
        arguments = remove_arg(arguments, '-isystem', 2, prefix=True)
        arguments = remove_arg(arguments, '-M', 1)
        arguments = remove_arg(arguments, '-MD', 1)
        arguments = remove_arg(arguments, '-MM', 1)
        arguments = remove_arg(arguments, '-MMD', 1)
        arguments = remove_arg(arguments, '-MF', 2, prefix=True)

        incs = [f'-I{dir}' for dir in vd.angle_bracket_include_paths]
        arguments[1:1] = ['-nostdinc', *incs]

    if deps_file.exists():
        deps_text = deps_file.read_text()
        deps_file.unlink()
        try:
            _target, deps = deps_text.split(':', 1)
        except Exception as e:
            print('failed to read dependencies:', deps_text, e)
            return index, set(), arguments
        lines = [l.rstrip('\\') for l in deps.splitlines()]
        files = shlex.split(' '.join(lines))

        dir_ = compilation_command.get('directory') or cc_dir

        return index, {Path(dir_, f) for f in files}, arguments
    else:
        print('no dependencies recorded for', compilation_command['file'])

    if completion.returncode != 0:
        print(completion.stderr)
    return index, set(), arguments



from shutil import copyfileobj
from socket import gethostname
from sys import exit
from urllib.parse import urlencode, urlparse
import http.server
import json
import os
import threading
import webbrowser

import requests
from platformdirs import user_config_path

from . import __version__


DEFAULT_UPLOAD_TOKEN_PATH = user_config_path('Territory') / 'upload_token'


def create_build_request(upload_token, repo_id, branch, meta, blob_size):
    uploader_api_url = _uploader_api_url()
    response = requests.post(
        uploader_api_url + '/build-request',
        json={
            'repo_id': repo_id,
            'branch': branch,
            'meta': meta,
            'len': blob_size,
        },
        headers={
            'Authorization': f'Bearer {upload_token}',
            'User-Agent': f'territory/{__version__}',
        })
    if not response.ok:
        print('HTTP status', response.status_code)
        exit(response.text)
    return response.json()


def download_resource(upload_token, resource, destination):
    uploader_api_url = _uploader_api_url()
    response = requests.get(
        uploader_api_url + '/build-resource/' + resource,
        headers={
            'Authorization': f'Bearer {upload_token}',
            'User-Agent': f'territory/{__version__}',
        })
    response.raise_for_status()
    destination.write_bytes(response.content)


def _uploader_api_url():
    return os.environ.get(
        'TERRITORY_UPLOAD_API',
        'https://maps.territory.dev')


def _authorizer_url():
    return os.environ.get(
        'TERRITORY_AUTHORIZER',
        'https://app.territory.dev/upload-tokens/authorize-local')


def _authorizer_origin():
    url = _authorizer_url()
    scheme, netloc, *_ = urlparse(url)
    origin = f'{scheme}://{netloc}'
    return origin


def open_authorizer(params):
    authorizer_url = _authorizer_url()
    url = authorizer_url + '?' + urlencode(params)
    webbrowser.open(url)


def auth(args):
    token_path = args.upload_token_path
    if token_path.is_file():
        print(f'reading token from {token_path}')
        return token_path.read_text().strip()
    else:
        token = _acquire_token()
        print(f'storing token in {token_path}')
        token_path.parent.mkdir(parents=True, exist_ok=True)
        token_path.touch(mode=0o600)
        token_path.write_text(token)
        return token


def _acquire_token():
    upload_token = None
    upload_token_received = threading.Event()

    class Handler(http.server.SimpleHTTPRequestHandler):
        def do_OPTIONS(self):
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.send_header('Access-Control-Allow-Origin', _authorizer_origin())
            self.end_headers()

        def do_POST(self):
            nonlocal upload_token, upload_token_received

            content_length = int(self.headers.get('Content-Length'), 10)
            body = self.rfile.read(content_length)

            try:
                upload_token = json.loads(body)['upload_token']
            except Exception:
                self.send_response(400)
                self.wfile.write(b"bad body")
                return

            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.send_header('Access-Control-Allow-Origin', _authorizer_origin())
            self.end_headers()
            self.wfile.write(b"OK")
            upload_token_received.set()


    print('We will start a local HTTP server and open the web browser now.')
    print('If a web browser is not available, create an upload token manually and')
    print(f'store it in {DEFAULT_UPLOAD_TOKEN_PATH}');
    s = http.server.HTTPServer(('localhost', 0), Handler)
    print('serving on', s.server_port)
    open_authorizer({
        'callback': f'http://localhost:{s.server_port}',
        'display_name': gethostname(),
    })

    thread = threading.Thread(target=s.serve_forever)
    thread.start()

    upload_token_received.wait()

    s.shutdown()
    return upload_token



from subprocess import check_output

from .files import find_in_ancestors



def find_repo_root(p):
    try:
        return find_in_ancestors(p, lambda p: (p / '.git').is_dir(), highest=True)
    except FileNotFoundError:
        raise SystemExit('not a git repository')


def list_repo_files(dir):
    return check_output(['git', 'ls-files'], cwd=dir, text=True).strip()


def get_branch(dir):
    return check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'], cwd=dir, text=True).strip()


def get_sha(dir) -> str:
    return check_output(['git', 'rev-parse', 'HEAD'], cwd=dir, text=True).strip()


def get_commit_message(dir) -> str:
    return check_output(['git', 'log', '-1', r'--format=%B'], cwd=dir, text=True)
